/*
   Source for x86 emulator IdaPro plugin
   File: emufuncs.cpp
   Copyright (c) 2004, Chris Eagle
   
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option) 
   any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
   FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
   more details.
   
   You should have received a copy of the GNU General Public License along with 
   this program; if not, write to the Free Software Foundation, Inc., 59 Temple 
   Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "x86defs.h"
#include "emufuncs.h"
#include "memmgr.h"
#include "hooklist.h"

#include <kernwin.hpp>
#include <bytes.hpp>
#include <name.hpp>

HookEntry hookTable[] = {
   {"VirtualAlloc", emu_VirtualAlloc},
   {"VirtualFree", emu_VirtualFree},
   {"GetProcAddress", emu_GetProcAddress},
   {"GetModuleHandle", emu_GetModuleHandle},
   {"LoadLibrary", emu_LoadLibrary},
   {"malloc", emu_malloc},
   {"calloc", emu_calloc},
   {"realloc", emu_realloc},
   {"free", emu_free},
   {NULL, NULL}
};


/*
 * Build a C string by reading from the specified address until a NULL is
 * encountered.  Returned value must be free'd
 *
 */

char *getString(MemoryManager *mgr, dword addr) {
   int size = 16;
   int i = 0;
   byte *str = NULL, ch;
   str = (byte*) malloc(size);
   while (ch = mgr->readByte(addr++)) {
      if (i == size) {
         str = (byte*)realloc(str, size + 16);
         size += 16;
      }
      str[i++] = ch;
   }
   if (i == size) {
      str = (byte*)realloc(str, size + 1);
   }
   str[i] = 0;
   return (char*)str;
}

/*
 * This function is used for all unemulated API functions
 */
void unemulated(MemoryManager *mgr, dword addr) {
   HookNode *n = find(addr);
   if (n) {
      msg("%s called without an emulation. Check your stack layout!\n", n->getName());
   }
}

/*
   These functions emulate various API calls.  The idea is
   to invoke them after all parameters have been pushed onto the
   stack.  Each function understands its corresponding parameters
   and calling conventions and leaves the stack in the proper state
   with a result in eax.  Because these are invoked from the emulator
   no return address gets pushed onto the stack and the functions can
   get right at their parameters on top of the stack.
*/

void emu_VirtualAlloc(MemoryManager *mgr, dword addr) {
   dword lpAddress = pop(SIZE_DWORD); 
   dword dwSize = pop(SIZE_DWORD);
   dword flAllocationType = pop(SIZE_DWORD);
   dword flProtect = pop(SIZE_DWORD);
   eax = mgr->heap->calloc(dwSize, 1);
}

void emu_VirtualFree(MemoryManager *mgr, dword addr) {
   mgr->heap->free(pop(SIZE_DWORD));
   dword dwSize = pop(SIZE_DWORD);
   dword dwFreeType = pop(SIZE_DWORD);
   eax = 1;    //determine actual return value for VirtualFree
}

/*
 * To do: Need to log all calls to GetProcAddress so that successive
 *        calls for same name can return same address.
 *        Need to account for "handles"??
 * Current behavior is a new "address" returned for each call.
 */

static char *lastProcName = NULL;

//FARPROC __stdcall GetProcAddress(HMODULE hModule,LPCSTR lpProcName)
void emu_GetProcAddress(MemoryManager *mgr, dword addr) {
   static dword address = 0;
   static dword bad = 0;
   dword hModule = pop(SIZE_DWORD); 
   dword lpProcName = pop(SIZE_DWORD);
   HookNode *n;
   int i;
   if (lastProcName == NULL) free(lastProcName);
   lastProcName = getString(mgr, lpProcName);
   msg("GetProcAddress called: %s", lastProcName);
   //first see if this function is already hooked
   if (n = find(lastProcName)) {
      eax = n->getAddr();
   }
   else {  //this is where we need to check if auto hooking is turned on else if (autohook) {
      //if it wasn't hooked, see if there is an emulation for it
      for (i = 0; hookTable[i].fName; i++) {
         if (!strcmp(hookTable[i].fName, lastProcName)) {
            //if there is an emulation, hook it
            addHook(lastProcName, ++address, hookTable[i].func);
            eax = address;
            break;
         }
      }
      if (hookTable[i].fName == NULL) {
         //there is no emulation, pass all calls to the "unemulated" stub
         addHook(lastProcName, --bad, unemulated);
         eax = bad;
      }
   }
   msg(" (0x%X)\n", eax);
}

/*
 * This is how we build import tables based on calls to 
 * GetProcAddress: create a label at addr from lastProcName.
 */

void makeImportLabel(dword addr) {
   for (dword cnt = 0; cnt < 4; cnt++) {
      do_unknown(addr, true); //undefine it
   }
   doDwrd(addr, 4);
   if (!set_name(addr, lastProcName, SN_PUBLIC | SN_NOCHECK | SN_NOWARN)) { //failed, probably duplicate name
      //undefine old name and retry once
      dword oldName = get_name_ea(BADADDR, lastProcName);
      if (oldName != BADADDR && del_global_name(oldName)) {
         set_name(addr, lastProcName, SN_PUBLIC | SN_NOCHECK | SN_NOWARN);
      }
   }
}

/*
 * To do: Need to log all calls to GetModuleHandle so that successive
 *        calls for same name can return same handle.  
 *        Need to integrate with LoadLibrary handles??
 * Current behavior is a new "handle" returned for each call.
 */

//HMODULE __stdcall GetModuleHandleA(LPCSTR lpModuleName)
void emu_GetModuleHandle(MemoryManager *mgr, dword addr) {
   static dword handle = 0;
   byte ch;
   dword lpModuleName = pop(SIZE_DWORD);
   msg("GetModuleHandle called: ");
   while (ch = mgr->readByte(lpModuleName++)) {
      msg("%c", ch);
   }
   eax = ++handle;
   msg(" (%d)\n", eax);
}

/*
 * To do: Need to log all calls to LoadLibrary so that successive
 *        calls for same name can return same handle.  
 *        Need to integrate with GetModuleHandle handles??
 * Current behavior is a new "handle" returned for each call.
 */

//HMODULE __stdcall LoadLibraryA(LPCSTR lpLibFileName)
void emu_LoadLibrary(MemoryManager *mgr, dword addr) {
   static dword handle = 0;
   byte ch;
   dword lpLibFileName = pop(SIZE_DWORD);
   msg("LoadLibrary called: ");
   while (ch = mgr->readByte(lpLibFileName++)) {
      msg("%c", ch);
   }
   eax = ++handle;
   msg(" (%d)\n", eax);
}

void emu_malloc(MemoryManager *mgr, dword addr) {
   eax = mgr->heap->malloc(readDword(esp));
}

void emu_calloc(MemoryManager *mgr, dword addr) {
   eax = mgr->heap->calloc(readDword(esp), readDword(esp + 4));
}

void emu_realloc(MemoryManager *mgr, dword addr) {
   eax = mgr->heap->realloc(readDword(esp), readDword(esp + 4));
}

void emu_free(MemoryManager *mgr, dword addr) {
   mgr->heap->free(pop(SIZE_DWORD));
}


